package org.openjdk.gcbench.util.ratelimit;

import java.util.concurrent.atomic.AtomicLongFieldUpdater;

public class TokenBucket implements RateLimiter {

    static final int QUANTA_PER_SEC = 5;
    static final int MS_PER_QUANTUM = 1000 / QUANTA_PER_SEC;

    static final AtomicLongFieldUpdater<TokenBucket> STATE =
            AtomicLongFieldUpdater.newUpdater(TokenBucket.class, "state");

    private final int tokensPerQuantum;
    private final long timeBase;
    private volatile long state;

    public TokenBucket(int ratePerSec) {
        this.tokensPerQuantum = Math.max(1, ratePerSec / QUANTA_PER_SEC);
        this.timeBase = System.currentTimeMillis();
    }

    private static int timestamp(long l) {
        return (int)(l >> 32);
    }

    private static int tokens(long l) {
        return (int)(l & 0x7FFFFFFF);
    }

    private static long pack(int timestamp, int tokens) {
        return ((long)timestamp << 32) + tokens;
    }

    @Override
    public void limit() {
        while (true) {
            int quantId = (int) ((System.currentTimeMillis() - timeBase) / MS_PER_QUANTUM);

            long cur = STATE.get(this);
            int time = timestamp(cur);
            int tokens = tokens(cur);

            if (time == quantId && tokens != 0) {
                // current quantum has tokens, try to claim and exit
                if (STATE.compareAndSet(this, cur, pack(quantId, tokens - 1))) {
                    return; // success
                } else {
                    continue; // immediate respin
                }
            } else if (time <= quantId) {
                // current or past quantum is empty, try to install a new one, and respin
                STATE.compareAndSet(this, cur, pack(quantId + 1, tokensPerQuantum));
            }

            // no rush: wait before respinning
            try {
                Thread.sleep(1);
            } catch (InterruptedException e) {
                // ignore
            }
        }
    }

}
